@@ -23,8 +23,12 @@
 
 #define _GNU_SOURCE
 #include <unistd.h>
+#include <fcntl.h>
+#include <stdlib.h>
+#include <limits.h>
 #include <sys/prctl.h>
 #include <sys/capability.h>
+#include <errno.h>
 
 #include "log.h"
 
@@ -105,8 +109,13 @@ int lxc_caps_up(void)
 
 		ret = cap_get_flag(caps, cap, CAP_PERMITTED, &flag);
 		if (ret) {
-			ERROR("failed to cap_get_flag: %m");
-			goto out;
+			if (errno == EINVAL) {
+				INFO("Last supported cap was %d\n", cap-1);
+				break;
+			} else {
+				ERROR("failed to cap_get_flag: %m");
+				goto out;
+			}
 		}
 
 		ret = cap_set_flag(caps, CAP_EFFECTIVE, 1, &cap, flag);
@@ -167,3 +176,85 @@ int lxc_caps_init(void)
 
 	return 0;
 }
+
+static int _real_caps_last_cap(void)
+{
+	int fd;
+	int result = -1;
+
+	/* try to get the maximum capability over the kernel
+	* interface introduced in v3.2 */
+	fd = open("/proc/sys/kernel/cap_last_cap", O_RDONLY);
+	if (fd >= 0) {
+		char buf[32];
+		char *ptr;
+		int n;
+
+		if ((n = read(fd, buf, 31)) >= 0) {
+			buf[n] = '\0';
+			result = strtol(buf, &ptr, 10);
+			if (!ptr || (*ptr != '\0' && *ptr != '\n') ||
+			    result == LONG_MIN || result == LONG_MAX)
+				result = -1;
+		}
+
+		close(fd);
+	}
+
+	/* try to get it manually by trying to get the status of
+	* each capability indiviually from the kernel */
+	if (result < 0) {
+		int cap = 0;
+		while (prctl(PR_CAPBSET_READ, cap) >= 0) cap++;
+		result = cap - 1;
+	}
+
+	return result;
+}
+
+int lxc_caps_last_cap(void)
+{
+	static int last_cap = -1;
+	if (last_cap < 0) last_cap = _real_caps_last_cap();
+
+	return last_cap;
+}
+
+/*
+ * check if we have the caps needed to start a container.  returns 1 on
+ * success, 0 on error.  (I'd prefer this be a bool, but am afraid that
+ * might fail to build on some distros).
+ */
+int lxc_caps_check(void)
+{
+	uid_t uid = getuid();
+	cap_t caps;
+	cap_flag_value_t value;
+	int i, ret;
+
+	cap_value_t needed_caps[] = { CAP_SYS_ADMIN, CAP_NET_ADMIN, CAP_SETUID, CAP_SETGID };
+
+#define NUMCAPS ((int) (sizeof(needed_caps) / sizeof(cap_t)))
+
+	if (!uid)
+		return 1;
+
+	caps = cap_get_proc();
+	if (!caps) {
+		ERROR("failed to cap_get_proc: %m");
+		return 0;
+	}
+
+	for (i=0; i<NUMCAPS; i++) {
+		ret = cap_get_flag(caps, needed_caps[i], CAP_EFFECTIVE, &value);
+		if (ret) {
+			ERROR("Failed to cap_get_flag: %m");
+			return 0;
+		}
+		if (!value) {
+			return 0;
+		}
+	}
+
+	return 1;
+}
