@@ -26,9 +26,10 @@
 #define MAXPRIOLEN 24
 
 struct lxc_handler;
-int lxc_cgroup_create(const char *name, pid_t pid);
-int lxc_cgroup_destroy(const char *name);
-int lxc_cgroup_path_get(char **path, const char *subsystem, const char *name);
-int lxc_cgroup_nrtasks(const char *name);
-int lxc_ns_is_mounted(void);
+extern int lxc_cgroup_create(const char *name, pid_t pid);
+extern int lxc_cgroup_destroy(const char *name);
+extern int lxc_cgroup_path_get(char **path, const char *subsystem, const char *name);
+extern int lxc_cgroup_nrtasks(const char *name);
+extern int lxc_cgroup_attach(const char *name, pid_t pid);
+extern int lxc_ns_is_mounted(void);
 #endif
