@@ -75,8 +75,14 @@ static int __lxc_command(const char *name, struct lxc_command *command,
 	int sock, ret = -1;
 	char path[sizeof(((struct sockaddr_un *)0)->sun_path)] = { 0 };
 	char *offset = &path[1];
+	int rc, len;
 
-	sprintf(offset, abstractname, name);
+	len = sizeof(path)-1;
+	rc = snprintf(offset, len, abstractname, name);
+	if (rc < 0 || rc >= len) {
+		ERROR("Name too long");
+		return -1;
+	}
 
 	sock = lxc_af_unix_connect(path);
 	if (sock < 0 && errno == ECONNREFUSED) {
@@ -131,10 +137,8 @@ pid_t get_init_pid(const char *name)
 	int ret, stopped = 0;
 
 	ret = lxc_command(name, &command, &stopped);
-	if (ret < 0 && stopped) {
-		ERROR("'%s' is not running", name);
+	if (ret < 0 && stopped)
 		return -1;
-	}
 
 	if (ret < 0) {
 		ERROR("failed to send command");
@@ -268,12 +272,23 @@ extern int lxc_command_mainloop_add(const char *name,
 	int ret, fd;
 	char path[sizeof(((struct sockaddr_un *)0)->sun_path)] = { 0 };
 	char *offset = &path[1];
+	int rc, len;
 
-	sprintf(offset, abstractname, name);
+	len = sizeof(path)-1;
+	rc = snprintf(offset, len, abstractname, name);
+	if (rc < 0 || rc >= len) {
+		ERROR("Name too long");
+		return -1;
+	}
 
 	fd = lxc_af_unix_open(path, SOCK_STREAM, 0);
 	if (fd < 0) {
-		ERROR("failed to create the command service point");
+		ERROR("failed (%d) to create the command service point %s", errno, offset);
+		if (errno == EADDRINUSE) {
+			ERROR("##");
+			ERROR("# The container appears to be already running!");
+			ERROR("##");
+		}
 		return -1;
 	}
 
