@@ -45,11 +45,15 @@ static int freeze_unfreeze(const char *name, int freeze)
 	char tmpf[32];
 	int fd, ret;
 	
-	ret = lxc_cgroup_path_get(&nsgroup, name);
+	ret = lxc_cgroup_path_get(&nsgroup, "freezer", name);
 	if (ret)
 		return -1;
 
-	snprintf(freezer, MAXPATHLEN, "%s/freezer.state", nsgroup);
+	ret = snprintf(freezer, MAXPATHLEN, "%s/freezer.state", nsgroup);
+	if (ret >= MAXPATHLEN) {
+		ERROR("freezer.state name too long");
+		return -1;
+	}
 
 	fd = open(freezer, O_RDWR);
 	if (fd < 0) {
